@@ -11,10 +11,16 @@ import {
   ReactWidget
 } from '@jupyterlab/apputils';
 import { Menu } from '@lumino/widgets';
+import { ReadonlyPartialJSONObject } from '@lumino/coreutils';
 import { ISettingRegistry } from '@jupyterlab/settingregistry';
 import { IStatusBar, TextItem } from '@jupyterlab/statusbar';
 import { Cell } from '@jupyterlab/cells';
 import { CodeMirrorEditor, ICodeMirror } from '@jupyterlab/codemirror';
+import {
+  ITranslator,
+  nullTranslator,
+  TranslationBundle
+} from '@jupyterlab/translation';
 
 import CodeMirror from 'codemirror';
 
@@ -40,11 +46,6 @@ const enum CommandIDs {
   chooseLanguage = 'spellchecker:choose-language'
 }
 
-const TEXT_SUGGESTIONS_AVAILABLE = 'Adjust spelling to';
-const TEXT_NO_SUGGESTIONS = 'No spellcheck suggestions';
-
-const PALETTE_CATEGORY = 'Spell Checker';
-
 interface IWord {
   line: number;
   start: number;
@@ -61,7 +62,7 @@ interface IContext {
  * Dictionary data defined by a pair of Hunspell .aff and .dic files.
  * More than one dictionary may exist for a given language.
  */
-interface IDictionary {
+interface IDictionary extends ReadonlyPartialJSONObject {
   /**
    * Identifier of the dictionary consisting of the filename without the .aff/.dic suffix
    * and path information if needed to distinguish from other dictionaries.
@@ -69,8 +70,9 @@ interface IDictionary {
   id: string;
   /**
    * BCP 47 code identifier.
+   * Absent for online dictionaries.
    */
-  code: string;
+  code?: string;
   /**
    * Display name, usually in the form "Language (Region)".
    */
@@ -83,15 +85,20 @@ interface IDictionary {
    * Path to the .dic file.
    */
   dic: string;
+  /**
+   * Indicated whether the dictionary is online.
+   */
+  isOnline: boolean;
 }
 
 interface ILanguageManagerResponse {
   version: string;
-  dictionaries: IDictionary[];
+  dictionaries: Omit<IDictionary, 'isOnline'>[];
 }
 
 class LanguageManager {
-  languages: IDictionary[];
+  protected serverDictionaries: IDictionary[];
+  protected onlineDictionaries: IDictionary[];
 
   public ready: Promise<any>;
 
@@ -99,40 +106,78 @@ class LanguageManager {
    * initialise the manager
    * mainly reading the definitions from the external extension
    */
-  constructor() {
-    // read the list of languages from the external extension
-    this.ready = requestAPI<any>('language_manager').then(
-      (values: ILanguageManagerResponse) => {
-        console.debug('LanguageManager is ready');
-        this.languages = values.dictionaries;
+  constructor(settingsRegistry: ISettingRegistry) {
+    const loadSettings = settingsRegistry.load(extension.id).then(settings => {
+      this.updateSettings(settings);
+      settings.changed.connect(() => {
+        this.updateSettings(settings);
+      });
+    });
+
+    this.ready = Promise.all([
+      this.fetchServerDictionariesList(),
+      loadSettings
+    ]).then(() => {
+      console.debug('LanguageManager is ready');
+    });
+  }
+
+  protected updateSettings(settings: ISettingRegistry.ISettings) {
+    if (settings) {
+      this.onlineDictionaries = (
+        settings.get('onlineDictionaries').composite as Omit<
+          IDictionary,
+          'isOnline' | 'code'
+        >[]
+      ).map(dictionary => {
+        return { ...dictionary, isOnline: true } as IDictionary;
+      });
+    }
+  }
+
+  /**
+   * Read the list of languages from the server extension
+   */
+  protected fetchServerDictionariesList(): Promise<void> {
+    return requestAPI<ILanguageManagerResponse>('language_manager').then(
+      values => {
+        console.debug('Dictionaries fetched from server');
+        this.serverDictionaries = values.dictionaries.map(dictionary => {
+          return {
+            ...dictionary,
+            isOnline: false
+          } as IDictionary;
+        });
       }
     );
   }
 
+  get dictionaries(): IDictionary[] {
+    return [...this.serverDictionaries, ...this.onlineDictionaries];
+  }
+
   /**
    * get an array of languages, put "language" in front of the list
    * the list is alphabetically sorted
    */
-  getChoices(language: IDictionary) {
-    return [
-      language,
-      ...this.languages
-        .filter(l => l.id !== language.id)
-        .sort((a, b) => a.name.localeCompare(b.name))
-    ];
+  getChoices(language: IDictionary | undefined) {
+    const options = language
+      ? [language, ...this.dictionaries.filter(l => l.id !== language.id)]
+      : this.dictionaries;
+    return options.sort((a, b) => a.name.localeCompare(b.name));
   }
 
   /**
    * select the language by the identifier
    */
   getLanguageByIdentifier(identifier: string): IDictionary | undefined {
-    const exactMatch = this.languages.find(l => l.id === identifier);
+    const exactMatch = this.dictionaries.find(l => l.id === identifier);
     if (exactMatch) {
       return exactMatch;
     }
     // approximate matches support transition from the 0.5 version (and older)
     // that used incorrect codes as language identifiers
-    const approximateMatch = this.languages.find(
+    const approximateMatch = this.dictionaries.find(
       l => l.id.toLowerCase() === identifier.replace('-', '_').toLowerCase()
     );
     if (approximateMatch) {
@@ -164,17 +209,21 @@ class SpellChecker {
   dictionary: any;
   suggestions_menu: Menu;
   status_widget: StatusWidget;
-  status_msg = 'Dictionary not loaded';
+  status_msg: string;
 
   // Default Options
   check_spelling = true;
-  language: IDictionary;
+  language: IDictionary | undefined;
   language_manager: LanguageManager;
   rx_word_char = /[^-[\]{}():/!;&@$£%§<>"*+=?.,~\\^|_`#±\s\t]/;
   rx_non_word_char = /[-[\]{}():/!;&@$£%§<>"*+=?.,~\\^|_`#±\s\t]/;
   ignored_tokens: Set<string> = new Set();
   settings: ISettingRegistry.ISettings;
   accepted_types: string[];
+  private _trans: TranslationBundle;
+  readonly TEXT_SUGGESTIONS_AVAILABLE: string;
+  readonly TEXT_NO_SUGGESTIONS: string;
+  readonly PALETTE_CATEGORY: string;
 
   constructor(
     protected app: JupyterFrontEnd,
@@ -182,11 +231,18 @@ class SpellChecker {
     protected editor_tracker: IEditorTracker,
     protected setting_registry: ISettingRegistry,
     protected code_mirror: ICodeMirror,
-    protected palette?: ICommandPalette,
-    protected status_bar?: IStatusBar
+    translator: ITranslator,
+    protected palette?: ICommandPalette | null,
+    protected status_bar?: IStatusBar | null
   ) {
     // use the language_manager
-    this.language_manager = new LanguageManager();
+    this.language_manager = new LanguageManager(setting_registry);
+    this._trans = translator.load('jupyterlab-spellchecker');
+
+    this.status_msg = this._trans.__('Dictionary not loaded');
+    this.TEXT_SUGGESTIONS_AVAILABLE = this._trans.__('Adjust spelling to');
+    this.TEXT_NO_SUGGESTIONS = this._trans.__('No spellcheck suggestions');
+    this.PALETTE_CATEGORY = this._trans.__('Spell Checker');
 
     // read the settings
     this.setup_settings();
@@ -315,7 +371,7 @@ class SpellChecker {
 
   setup_button() {
     this.app.commands.addCommand(CommandIDs.toggle, {
-      label: 'Toggle spellchecker',
+      label: this._trans.__('Toggle spellchecker'),
       execute: () => {
         this.toggle_spellcheck();
       }
@@ -323,7 +379,7 @@ class SpellChecker {
     if (this.palette) {
       this.palette.addItem({
         command: CommandIDs.toggle,
-        category: PALETTE_CATEGORY
+        category: this.PALETTE_CATEGORY
       });
     }
   }
@@ -375,7 +431,7 @@ class SpellChecker {
 
   setup_suggestions() {
     this.suggestions_menu = new Menu({ commands: this.app.commands });
-    this.suggestions_menu.title.label = TEXT_SUGGESTIONS_AVAILABLE;
+    this.suggestions_menu.title.label = this.TEXT_SUGGESTIONS_AVAILABLE;
     this.suggestions_menu.title.icon = spellcheckIcon.bindprops({
       stylesheet: 'menuItem'
     });
@@ -414,7 +470,7 @@ class SpellChecker {
       execute: () => {
         this.ignore();
       },
-      label: 'Ignore'
+      label: this._trans.__('Ignore')
     });
 
     this.app.contextMenu.addItem({
@@ -461,12 +517,12 @@ class SpellChecker {
           args: { name: suggestion }
         });
       }
-      this.suggestions_menu.title.label = TEXT_SUGGESTIONS_AVAILABLE;
+      this.suggestions_menu.title.label = this.TEXT_SUGGESTIONS_AVAILABLE;
       this.suggestions_menu.title.className = '';
       this.suggestions_menu.setHidden(false);
     } else {
       this.suggestions_menu.title.className = 'lm-mod-disabled';
-      this.suggestions_menu.title.label = TEXT_NO_SUGGESTIONS;
+      this.suggestions_menu.title.label = this.TEXT_NO_SUGGESTIONS;
       this.suggestions_menu.setHidden(true);
     }
   }
@@ -495,16 +551,22 @@ class SpellChecker {
   }
 
   load_dictionary() {
-    this.status_msg = 'Loading dictionary ...';
+    const language = this.language;
+    if (!language) {
+      return new Promise((accept, reject) =>
+        reject('Cannot load dictionary: no language set')
+      );
+    }
+    this.status_msg = this._trans.__('Loading dictionary…');
     this.status_widget.update();
     return Promise.all([
-      fetch(this.language.aff).then(res => res.text()),
-      fetch(this.language.dic).then(res => res.text())
+      fetch(language.aff).then(res => res.text()),
+      fetch(language.dic).then(res => res.text())
     ]).then(values => {
-      this.dictionary = new Typo(this.language.name, values[0], values[1]);
-      console.debug('Dictionary Loaded ', this.language.name, this.language.id);
+      this.dictionary = new Typo(language.name, values[0], values[1]);
+      console.debug('Dictionary Loaded ', language.name, language.id);
 
-      this.status_msg = this.language.name;
+      this.status_msg = language.name;
       // update the complete UI
       this.status_widget.update();
       this.refresh_state();
@@ -563,11 +625,14 @@ class SpellChecker {
     const choiceStrings = choices.map(
       // note: two dictionaries may exist for a language with the same name,
       // so we append the actual id of the dictionary in the square brackets.
-      dictionary => dictionary.name + ' [' + dictionary.id + ']'
+      dictionary =>
+        dictionary.isOnline
+          ? this._trans.__('%1 [%2] (online)', dictionary.name, dictionary.id)
+          : this._trans.__('%1 [%2]', dictionary.name, dictionary.id)
     );
 
     InputDialog.getItem({
-      title: 'Choose spellchecker language',
+      title: this._trans.__('Choose spellchecker language'),
       items: choiceStrings
     }).then(value => {
       if (value.value !== null) {
@@ -595,13 +660,13 @@ class SpellChecker {
     };
     this.app.commands.addCommand(CommandIDs.chooseLanguage, {
       execute: args => this.choose_language(),
-      label: 'Choose spellchecker language'
+      label: this._trans.__('Choose spellchecker language')
     });
 
     if (this.palette) {
       this.palette.addItem({
         command: CommandIDs.chooseLanguage,
-        category: PALETTE_CATEGORY
+        category: this.PALETTE_CATEGORY
       });
     }
 
@@ -623,8 +688,9 @@ function activate(
   editor_tracker: IEditorTracker,
   setting_registry: ISettingRegistry,
   code_mirror: ICodeMirror,
-  palette: ICommandPalette,
-  status_bar: IStatusBar
+  translator: ITranslator | null,
+  palette: ICommandPalette | null,
+  status_bar: IStatusBar | null
 ): void {
   console.log('Attempting to load spellchecker');
   const sp = new SpellChecker(
@@ -633,6 +699,7 @@ function activate(
     editor_tracker,
     setting_registry,
     code_mirror,
+    translator || nullTranslator,
     palette,
     status_bar
   );
@@ -646,7 +713,7 @@ const extension: JupyterFrontEndPlugin<void> = {
   id: '@ijmbarr/jupyterlab_spellchecker:plugin',
   autoStart: true,
   requires: [INotebookTracker, IEditorTracker, ISettingRegistry, ICodeMirror],
-  optional: [ICommandPalette, IStatusBar],
+  optional: [ITranslator, ICommandPalette, IStatusBar],
   activate: activate
 };
 
